/*
 * Copyright (C) 2017 by
 *   MetraLabs GmbH (MLAB), GERMANY
 * and
 *   Neuroinformatics and Cognitive Robotics Labs (NICR) at TU Ilmenau, GERMANY
 * All rights reserved.
 *
 * Contact: info@mira-project.org
 *
 * Commercial Usage:
 *   Licensees holding valid commercial licenses may use this file in
 *   accordance with the commercial license agreement provided with the
 *   software or, alternatively, in accordance with the terms contained in
 *   a written agreement between you and MLAB or NICR.
 *
 * GNU General Public License Usage:
 *   Alternatively, this file may be used under the terms of the GNU
 *   General Public License version 3.0 as published by the Free Software
 *   Foundation and appearing in the file LICENSE.GPL3 included in the
 *   packaging of this file. Please review the following information to
 *   ensure the GNU General Public License version 3.0 requirements will be
 *   met: http://www.gnu.org/copyleft/gpl.html.
 *   Alternatively you may (at your option) use any later version of the GNU
 *   General Public License if such license has been publicly approved by
 *   MLAB and NICR (or its successors, if any).
 *
 * IN NO EVENT SHALL "MLAB" OR "NICR" BE LIABLE TO ANY PARTY FOR DIRECT,
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
 * THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF "MLAB" OR
 * "NICR" HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * "MLAB" AND "NICR" SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND "MLAB" AND "NICR" HAVE NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR MODIFICATIONS.
 */

/**
 * @file SceneQueryTool.C
 *    Implementation of SceneQueryTool.h.
 *
 * @author Christof Schröter
 * @date   2017/02/07
 */

#include <visualization/3d/SceneQueryTool.h>

#include <OGRE/OgreSceneManager.h>

#include <serialization/DefaultInitializer.h>

namespace mira {

///////////////////////////////////////////////////////////////////////////////

SceneQueryTool::SceneQueryTool() : mDevicePixelRatio(1.f)
{
	MIRA_INITIALIZE_THIS;
}

void SceneQueryTool::setupScene(IVisualization3DSite* site)
{
	VisualizationTool3D::setupScene(site);
	mRaySceneQuery = site->getSceneManager()->createRayQuery(Ogre::Ray());
}

std::pair<std::string, Ogre::Vector3> SceneQueryTool::getIntersection(float x, float y,
                                                                      const std::string& filter,
                                                                      bool exactMatch,
                                                                      bool transform)
{
	auto view = getSite()->getSceneManager()->getCurrentViewport();
	auto camera = view->getCamera();
	Ogre::Ray mouseRay = camera->getCameraToViewportRay(x*mDevicePixelRatio/view->getActualWidth(),
	                                                    y*mDevicePixelRatio/view->getActualHeight());
	mRaySceneQuery->setRay(mouseRay);

	if (mBackgroundMode != IGNORE_OBJECTS) {
		Ogre::RaySceneQueryResult& result = mRaySceneQuery->execute();
		foreach(const auto& i, result)
		{
			if(!i.movable)
				continue;

			if (filter.empty())
				return std::make_pair(i.movable->getName(),
									  transform ? transformVector(mouseRay.getPoint(i.distance))
												: mouseRay.getPoint(i.distance));

			if (!exactMatch && (i.movable->getName().find(filter) != std::string::npos))
				return std::make_pair(i.movable->getName(),
									  transform ? transformVector(mouseRay.getPoint(i.distance))
												: mouseRay.getPoint(i.distance));

			if (i.movable->getName() == filter)
				return std::make_pair(i.movable->getName(),
									  transform ? transformVector(mouseRay.getPoint(i.distance))
												: mouseRay.getPoint(i.distance));
		}
	}

	if (mBackgroundMode == IGNORE_BACKGROUND)
		return std::make_pair("", Ogre::Vector3());

	return std::make_pair("-Background-", mouseRay.getPoint(camera->getPosition().z - mBackgroundHeight));
}

void SceneQueryTool::onMousePressed(QMouseEvent* e, float devicePixelRatio)
{
	mDevicePixelRatio = devicePixelRatio;
	((VisualizationTool*)this)->onMousePressed(e);
}

void SceneQueryTool::onMouseReleased(QMouseEvent* e, float devicePixelRatio)
{
	mDevicePixelRatio = devicePixelRatio;
	((VisualizationTool*)this)->onMouseReleased(e);
}

void SceneQueryTool::onMouseDoubleClick(QMouseEvent* e, float devicePixelRatio)
{
	mDevicePixelRatio = devicePixelRatio;
	((VisualizationTool*)this)->onMouseDoubleClick(e);
}

void SceneQueryTool::onMouseMoved(QMouseEvent* e, float devicePixelRatio)
{
	mDevicePixelRatio = devicePixelRatio;
	((VisualizationTool*)this)->onMouseMoved(e);
}

void SceneQueryTool::onMouseWheel(QWheelEvent * e, float devicePixelRatio)
{
	mDevicePixelRatio = devicePixelRatio;
	((VisualizationTool*)this)->onMouseWheel(e);
}


///////////////////////////////////////////////////////////////////////////////

} // namespace
